
from dataclasses import dataclass, field
import inspect
import math
from typing import List, Optional, Tuple
import mlx.nn as nn
import mlx.core as mx

@dataclass
class MBartConfig:
    model_type: str = "mbart"
    vocab_size: int = 50265
    d_model: int = 1024
    encoder_layers: int = 12
    decoder_layers: int = 12
    encoder_attention_heads: int = 16
    decoder_attention_heads: int = 16
    decoder_ffn_dim: int = 4096
    encoder_ffn_dim: int = 4096
    activation_function: str = "gelu"
    dropout: float = 0.1
    attention_dropout: float = 0.0
    activation_dropout: float = 0.0
    classifier_dropout: float = 0.0
    max_position_embeddings: int = 1024
    init_std: float = 0.02
    encoder_layerdrop: float = 0.0
    decoder_layerdrop: float = 0.0
    scale_embedding: bool = False
    use_cache: bool = True
    forced_eos_token_id: int = 2
    pad_token_id: int = 1
    bos_token_id: int = 0
    eos_token_id: int = 2
    is_encoder_decoder: bool = True
    keys_to_ignore_at_inference: List[str] = field(default_factory=lambda: ["past_key_values"])
    attribute_map: dict = field(default_factory=lambda: {"num_attention_heads": "encoder_attention_heads", "hidden_size": "d_model"})

    def __post_init__(self):
        self.num_hidden_layers = self.encoder_layers
        
    @classmethod
    def from_dict(cls, params):
        return cls(
            **{
                k: v
                for k, v in params.items()
                if k in inspect.signature(cls).parameters
            }
        )
class MBartAttention(nn.Module):
    def __init__(self, config: MBartConfig):
        super().__init__()
        self.num_heads = config.encoder_attention_heads
        self.embed_dim = config.d_model
        self.head_dim = self.embed_dim // self.num_heads
        self.scaling = self.head_dim**-0.5
        self.k_proj = nn.Linear(self.embed_dim, self.embed_dim, bias=True)
        self.v_proj = nn.Linear(self.embed_dim, self.embed_dim, bias=True)
        self.q_proj = nn.Linear(self.embed_dim, self.embed_dim, bias=True)
        self.out_proj = nn.Linear(self.embed_dim, self.embed_dim, bias=True)

    def _shape(self, proj:mx.array, seq_len: int, bsz: int):
        # hardcode as bs=1, seqlen=1
        return proj.reshape(bsz, seq_len, self.num_heads, self.head_dim).transpose(
            0, 2, 1, 3
        )
    def __call__(self, hidden_states: mx.array, past_key_value: Optional[Tuple[Tuple[mx.array]]]  = None, encoder_hidden_states: Optional[mx.array] = None,
        ) -> mx.array:
        is_cross_attention = encoder_hidden_states is not None

        bsz, tgt_len, _ = hidden_states.shape

        # get query proj
        query_states = self.q_proj(hidden_states)
        # get key, value proj
        # `past_key_value[0].shape[2] == key_value_states.shape[1]`
        # is checking that the `sequence_length` of the `past_key_value` is the same as
        # the provided `key_value_states` to support prefix tuning
        if (
            is_cross_attention
            and past_key_value is not None
            and past_key_value[0].shape[2] == encoder_hidden_states.shape[1]
        ):
            # reuse k,v, cross_attentions
            key_states = past_key_value[0]
            value_states = past_key_value[1]
        elif is_cross_attention:
            # cross_attentions
            key_states = self._shape(self.k_proj(encoder_hidden_states), -1, bsz)
            value_states = self._shape(self.v_proj(encoder_hidden_states), -1, bsz)
        elif past_key_value is not None:
            # reuse k, v, self_attention
            key_cache, value_cache = past_key_value
            key_states = self._shape(self.k_proj(hidden_states), -1, bsz)
            value_states = self._shape(self.v_proj(hidden_states), -1, bsz)
            key_states = mx.concatenate([key_cache, key_states], axis=2)
            value_states = mx.concatenate([value_cache, value_states], axis=2)
        else:
            # self_attention
            key_states = self._shape(self.k_proj(hidden_states), -1, bsz)
            value_states = self._shape(self.v_proj(hidden_states), -1, bsz)

        past_key_value = (key_states, value_states)

        query_states = self._shape(query_states, tgt_len, bsz)

        attn_output = mx.fast.scaled_dot_product_attention(
            query_states, key_states, value_states, scale=self.scaling
        )

        attn_output = attn_output.transpose(0, 2, 1, 3)

        # Use the `embed_dim` from the config (stored in the class) rather than `hidden_state` because `attn_output` can be
        # partitioned across GPUs when using tensor-parallelism.
        attn_output = attn_output.reshape(bsz, tgt_len, self.embed_dim)

        attn_output = self.out_proj(attn_output)

        return attn_output, past_key_value
    
class MBartDecoderLayer(nn.Module):
    def __init__(self, config: MBartConfig):
            super().__init__()
            self.embed_dim = config.d_model
            self.self_attn = MBartAttention(config)
            self.encoder_attn = MBartAttention(config)
            self.self_attn_layer_norm = nn.LayerNorm(self.embed_dim)
            self.encoder_attn_layer_norm = nn.LayerNorm(self.embed_dim)
            self.activation_fn = nn.gelu
            self.fc1 = nn.Linear(self.embed_dim, config.decoder_ffn_dim)
            self.fc2 = nn.Linear(config.decoder_ffn_dim, self.embed_dim)
            self.final_layer_norm = nn.LayerNorm(self.embed_dim)

    def __call__(self, hidden_states: mx.array, past_key_value: Optional[Tuple[Tuple[mx.array]]] = None, encoder_hidden_states: Optional[mx.array] = None,
        ) -> mx.array:        
        residual = hidden_states
        hidden_states = self.self_attn_layer_norm(hidden_states)
        self_attn_past_key_value = past_key_value[:2] if past_key_value is not None else None
    
        hidden_states, present_key_value = self.self_attn(
                    hidden_states=hidden_states,
                    past_key_value=self_attn_past_key_value,
                )
        hidden_states = residual + hidden_states

        residual = hidden_states
        hidden_states = self.encoder_attn_layer_norm(hidden_states)
        cross_attn_past_key_value = past_key_value[-2:] if past_key_value is not None else None
        hidden_states, cross_attn_present_key_value = self.encoder_attn(
            hidden_states=hidden_states,
            encoder_hidden_states=encoder_hidden_states,
            past_key_value=cross_attn_past_key_value,
        )
        hidden_states = residual + hidden_states
        present_key_value = present_key_value + cross_attn_present_key_value
        residual = hidden_states
        hidden_states = self.final_layer_norm(hidden_states)
        hidden_states = self.activation_fn(self.fc1(hidden_states))
        hidden_states = self.fc2(hidden_states)
        hidden_states = residual + hidden_states

        return hidden_states, present_key_value
    
class MBartDecoder(nn.Module):
    def __init__(self, config: MBartConfig):
        super().__init__()
        self.embed_scale = math.sqrt(config.d_model)

        self.embed_tokens = nn.Embedding(config.vocab_size, config.d_model)
        self.embed_positions = nn.Embedding(
                    config.max_position_embeddings+2,
                    config.d_model,
                )
        self.layers = [MBartDecoderLayer(config) for _ in range(config.decoder_layers)]
        self.layernorm_embedding = nn.LayerNorm(config.d_model)
        self.layer_norm = nn.LayerNorm(config.d_model)            

    def __call__(self, input_ids: mx.array, past_key_value: Optional[Tuple[Tuple[mx.array]]] = None, encoder_hidden_states: Optional[mx.array] = None,
        ) -> mx.array:
            past_key_values_length = past_key_value[0][0].shape[2] if past_key_value is not None else 0
            inputs_embeds = self.embed_tokens(input_ids) * self.embed_scale
            positions = self.embed_positions(mx.array(past_key_values_length + 2))
            hidden_states = inputs_embeds + positions
            hidden_states = self.layernorm_embedding(hidden_states)

            hidden_states = hidden_states.reshape(1, 1, -1)

            if past_key_value is None:
                past_key_value = [None] * len(self.layers)

            for idx, decoder_layer in enumerate(self.layers):
                hidden_states, past_key_value[idx] = decoder_layer(
                    hidden_states,
                    encoder_hidden_states=encoder_hidden_states,
                    past_key_value=past_key_value[idx],
                )

            hidden_states = self.layer_norm(hidden_states)

            return hidden_states, past_key_value

class MBartModel(nn.Module):
    def __init__(self, config: MBartConfig):
        super().__init__()
        self.model = MBartDecoder(config)
        self.lm_head = nn.Linear(config.d_model, config.vocab_size, bias=False)
    def __call__(self, x: mx.array, past_key_value: Optional[Tuple[Tuple[mx.array]]] = None, encoder_hidden_states: Optional[mx.array] = None,
        ) -> mx.array:   
        x, past_key_value = self.model(x, past_key_value, encoder_hidden_states)
        x = self.lm_head(x)
        return x, past_key_value
    
    @staticmethod
    def sanitize(weights):
        sanitized_weights = {}
        for k, v in weights.items():
            if k.startswith("encoder."):
                continue
            # remove `encoder.` prefix in k
            k = k.replace("decoder.", "")
            sanitized_weights[k] = v
        return sanitized_weights


#!/usr/bin/env python3

import argparse
import time
from pathlib import Path
from typing import Any

import numpy as np
import requests
import mlx.core as mx
from PIL import Image as PILImage
from transformers import NougatProcessor
from wand.image import Image as WandImage
from wand.color import Color

from mlx_nougat.nougat import Nougat

def download_file(url, local_filename):
    with requests.get(url, stream=True) as r:
        r.raise_for_status()
        with open(local_filename, 'wb') as f:
            for chunk in r.iter_content(chunk_size=8192):
                f.write(chunk)
    return local_filename

def extract_pdf_pages_as_images(filename: str, *, resolution=200):
    assert filename.endswith(
        ".pdf"
    ), f"{filename} does not end with .pdf, is it a pdf file?"
    print(f"start extracting {filename}")
    # Load the image using wand
    images = []
    with WandImage(filename=filename, resolution=resolution) as wand_imgs:
        print(f"the pdf has {len(wand_imgs.sequence)} pages")
        for wand_img in wand_imgs.sequence:
            wand_img.background_color = Color("white")
            wand_img.alpha_channel = "remove"
            numpy_array = np.array(wand_img)
            images.append(PILImage.fromarray(numpy_array).convert("RGB"))
    return images

def load_image(image_source):
    if image_source.startswith(("http://", "https://")):
        try:
            response = requests.get(image_source, stream=True)
            response.raise_for_status()
            img = PILImage.open(response.raw)
        except Exception as e:
            raise ValueError(
                f"Failed to load image from URL: {image_source} with error {e}"
            )
    elif Path(image_source).is_file():
        try:
            img = PILImage.open(image_source)
        except IOError as e:
            raise ValueError(f"Failed to load image {image_source} with error: {e}")
    else:
        raise ValueError(
            f"The image {image_source} must be a valid URL or existing file."
        )
    
    img = img.convert('RGB')
    return img

def apply_repetition_penalty(logits: mx.array, generated_tokens: Any, penalty: float):
    if len(generated_tokens) > 0:
        indices = mx.array([token for token in generated_tokens])
        selected_logits = logits[:, indices]
        selected_logits = mx.where(
            selected_logits < 0, selected_logits * penalty, selected_logits / penalty
        )
        logits[:, indices] = selected_logits
    return logits   

def top_p_sampling(logits: mx.array, top_p: float, temperature: float) -> mx.array:
    probs = mx.softmax(logits / temperature, axis=-1)

    sorted_indices = mx.argsort(probs, axis=-1)
    sorted_probs = probs[..., sorted_indices.squeeze(0)]

    cumulative_probs = mx.cumsum(sorted_probs, axis=-1)

    top_probs = mx.where(
        cumulative_probs > 1 - top_p,
        sorted_probs,
        mx.zeros_like(sorted_probs),
    )

    sorted_token = mx.random.categorical(mx.log(top_probs))
    token = sorted_indices.squeeze(0)[sorted_token]

    return token

def sample(logits, temperature=0.0, top_p=0.0):
    if temperature == 0:
        return mx.argmax(logits, axis=-1)
    else:
        return top_p_sampling(logits, top_p, temperature) if top_p > 0 and top_p < 1 else mx.random.categorical(logits * (1 / temperature))
    
def generate(model, pixel_values, max_new_tokens=4096, eos_token_id=2, temperature=0.0, top_p=0.0, repetition_penalty=None, repetition_context_size=10):
    encoder_hidden_states = model.encoder(pixel_values)
    new_token = mx.array([0])
    outputs = []
    cache = None
    repetition_context = []
    if repetition_context_size:
        repetition_context = repetition_context[-repetition_context_size:]
    
    for _ in range(max_new_tokens):
        logits, cache = model.decoder(new_token, cache, encoder_hidden_states)
        logits = logits[:, -1, :]

        if repetition_penalty:
            logits = apply_repetition_penalty(
                logits, repetition_context, repetition_penalty
            )
        
        new_token = sample(logits, temperature, top_p)
        
        if repetition_penalty:
            repetition_context.append(new_token.item())
        
        if new_token == eos_token_id:
            break
        outputs.append(new_token.item())
    return outputs

def main():
    parser = argparse.ArgumentParser(description="OCR tool using Nougat model")
    parser.add_argument("--model", default="facebook/nougat-small", help="Model name or path")
    parser.add_argument("--input", required=True, help="Input image or PDF file path or URL")
    parser.add_argument("--output", help="Output file path to save results")
    parser.add_argument("--temperature", type=float, default=0.0, help="Sampling temperature")
    parser.add_argument("--top_p", type=float, default=0.0, help="Top-p sampling threshold")
    parser.add_argument("--repetition_penalty", type=float, default=None, help="Repetition penalty")
    args = parser.parse_args()

    print(f"Loading model: {args.model}")
    nougat_processor = NougatProcessor.from_pretrained(args.model)
    model = Nougat.from_pretrained(args.model)

    # Handle remote PDF
    if args.input.lower().startswith(("http://", "https://")) and args.input.lower().endswith('.pdf'):
        print(f"Downloading PDF from {args.input}")
        local_filename = "downloaded_file.pdf"
        download_file(args.input, local_filename)
        args.input = local_filename

    if args.input.lower().endswith('.pdf'):
        images = extract_pdf_pages_as_images(args.input)
    else:
        images = [load_image(args.input)]

    results = []
    start_time = time.time()

    for i, img in enumerate(images):
        print(f"Processing page {i+1}/{len(images)}")
        pixel_values = mx.array(nougat_processor(img, return_tensors="np").pixel_values).transpose(0, 2, 3, 1)
        outputs = generate(model, pixel_values, max_new_tokens=4096, 
                           eos_token_id=nougat_processor.tokenizer.eos_token_id,
                           temperature=args.temperature,
                           top_p=args.top_p,
                           repetition_penalty=args.repetition_penalty)
        results.append(nougat_processor.tokenizer.decode(outputs))

    end_time = time.time()
    elapsed_time = end_time - start_time

    output_text = "\n\n".join(results)
    print(output_text)
    print(f"\nGeneration time: {elapsed_time:.2f} seconds")

    if args.output:
        with open(args.output, 'w', encoding='utf-8') as f:
            f.write(output_text)
        print(f"Results saved to {args.output}")

if __name__ == "__main__":
    main()


import json
from pathlib import Path
from typing import Any, Dict, Optional, Union
import mlx.core as mx
from huggingface_hub import snapshot_download
from huggingface_hub.utils._errors import RepositoryNotFoundError

def custom_roll(arr, shift, axis):
    if not isinstance(axis, tuple):
        axis = (axis,)
    if not isinstance(shift, tuple):
        shift = (shift,)
    
    for ax, sh in zip(axis, shift):
        arr = custom_roll_single_axis(arr, sh, ax)
    
    return arr

def custom_roll_single_axis(arr, shift, axis):
    if shift == 0:
        return arr
    
    shape = arr.shape
    n = shape[axis]
    
    shift = shift % n 
    
    indices = mx.concatenate((mx.arange(n - shift, n), mx.arange(n - shift)))
    
    return mx.take(arr, indices, axis=axis)


def window_partition(input_feature, window_size):
    batch_size, height, width, num_channels = input_feature.shape
    input_feature = input_feature.reshape(
        batch_size, height // window_size, window_size, width // window_size, window_size, num_channels
    )
    windows = input_feature.transpose(0, 1, 3, 2, 4, 5).reshape(-1, window_size, window_size, num_channels)
    return windows


def window_reverse(windows, window_size, height, width):
    num_channels = windows.shape[-1]
    windows = windows.reshape(-1, height // window_size, width // window_size, window_size, window_size, num_channels)
    windows = windows.transpose(0, 1, 3, 2, 4, 5).reshape(-1, height, width, num_channels)
    return windows

class ModelNotFoundError(Exception):
    def __init__(self, message):
        self.message = message
        super().__init__(self.message)

def get_model_path(path_or_hf_repo: str, revision: Optional[str] = None) -> Path:
    model_path = Path(path_or_hf_repo)
    if not model_path.exists():
        try:
            model_path = Path(
                snapshot_download(
                    repo_id=path_or_hf_repo,
                    revision=revision,
                    allow_patterns=[
                        "*.json",
                        "*.safetensors",
                        "*.py",
                        "tokenizer.model",
                        "*.tiktoken",
                        "*.txt",
                    ],
                )
            )
        except RepositoryNotFoundError:
            raise ModelNotFoundError(
                f"Model not found for path or HF repo: {path_or_hf_repo}.\n"
                "Please make sure you specified the local path or Hugging Face"
                " repo id correctly.\nIf you are trying to access a private or"
                " gated Hugging Face repo, make sure you are authenticated:\n"
                "https://huggingface.co/docs/huggingface_hub/en/guides/cli#huggingface-cli-login"
            ) from None
    return model_path

def make_shards(weights: dict, max_file_size_gb: int = 5) -> list:
    """
    Splits the weights into smaller shards.

    Args:
        weights (dict): Model weights.
        max_file_size_gb (int): Maximum size of each shard in gigabytes.

    Returns:
        list: List of weight shards.
    """
    max_file_size_bytes = max_file_size_gb << 30
    shards = []
    shard, shard_size = {}, 0
    for k, v in weights.items():
        if shard_size + v.nbytes > max_file_size_bytes:
            shards.append(shard)
            shard, shard_size = {}, 0
        shard[k] = v
        shard_size += v.nbytes
    shards.append(shard)
    return shards

def save_weights(
    save_path: Union[str, Path],
    weights: Dict[str, Any],
    *,
    donate_weights: bool = False,
) -> None:
    """Save model weights into specified directory."""
    if isinstance(save_path, str):
        save_path = Path(save_path)
    save_path.mkdir(parents=True, exist_ok=True)

    shards = make_shards(weights)
    shards_count = len(shards)
    shard_file_format = (
        "model-{:05d}-of-{:05d}.safetensors"
        if shards_count > 1
        else "model.safetensors"
    )

    total_size = sum(v.nbytes for v in weights.values())
    index_data = {"metadata": {"total_size": total_size}, "weight_map": {}}

    # Write the weights and make sure no references are kept other than the
    # necessary ones
    if donate_weights:
        weights.clear()
        del weights

    for i in range(len(shards)):
        shard = shards[i]
        shards[i] = None
        shard_name = shard_file_format.format(i + 1, shards_count)
        shard_path = save_path / shard_name

        mx.save_safetensors(str(shard_path), shard, metadata={"format": "mlx"})

        for weight_name in shard.keys():
            index_data["weight_map"][weight_name] = shard_name
        del shard

    index_data["weight_map"] = {
        k: index_data["weight_map"][k] for k in sorted(index_data["weight_map"])
    }

    with open(save_path / "model.safetensors.index.json", "w") as f:
        json.dump(
            index_data,
            f,
            indent=4,
        )


def save_config(
    config: dict,
    config_path: Union[str, Path],
) -> None:
    # Clean unused keys
    config.pop("_name_or_path", None)

    # sort the config for better readability
    config = dict(sorted(config.items()))
    # write the updated config to the config_path (if provided)
    with open(config_path, "w") as fid:
        json.dump(config, fid, indent=4)


import glob
import json
from pathlib import Path
from huggingface_hub import snapshot_download

import mlx.core as mx
import mlx.nn as nn

from mlx_nougat.mbartModel import MBartConfig, MBartModel
from mlx_nougat.donutSwinModel import DonutSwinModel, DonutSwinModelConfig
from mlx_nougat.utils import get_model_path

class Nougat(nn.Module):
    def __init__(self):
        super().__init__()
        self.encoder = None
        self.decoder = None
    @staticmethod
    def from_pretrained(path_or_hf_repo: str):
        path = get_model_path(path_or_hf_repo)

        with open(path / "config.json", "r") as f:
            model_config = json.load(f)

        model = Nougat()
        encoder_config = DonutSwinModelConfig.from_dict(model_config['encoder'])
        deconder_config = MBartConfig.from_dict(model_config['decoder'])
        donutSwinModel =DonutSwinModel(encoder_config)
        mBartModel = MBartModel(deconder_config)
        weight_files = glob.glob(str(path / "*.safetensors"))
        if not weight_files:
            raise FileNotFoundError(f"No safetensors found in {path}")

        weights = {}
        for wf in weight_files:
            weights.update(mx.load(wf))

        model.encoder = donutSwinModel
        model.decoder = mBartModel


        if (quantization := model_config.get("quantization", None)) is not None:
            def class_predicate(p, m):
                return f"{p}.scales" in weights

            nn.quantize(
                model,
                **quantization,
                class_predicate=class_predicate,
            )
            donutSwinModel_weights =  {k.replace("encoder.", "", 1): v for k, v in weights.items() if not k.startswith("decoder.")}
            mBartModel_weights = {k.replace("decoder.", ""): v for k, v in weights.items() if not k.startswith("encoder.")}
        else:
            donutSwinModel_weights = DonutSwinModel.sanitize(weights)
            mBartModel_weights = MBartModel.sanitize(weights)

        donutSwinModel.load_weights(list(donutSwinModel_weights.items()))
        mBartModel.load_weights(list(mBartModel_weights.items()))

        return model        



import json
from pathlib import Path
import shutil
from typing import Tuple
from mlx.utils import tree_flatten
import mlx.core as mx
import mlx.nn as nn
from mlx_nougat.nougat import Nougat
from mlx_nougat.utils import get_model_path, save_config, save_weights
import copy


def quantize_model(
    model: nn.Module, config: dict, q_group_size: int, q_bits: int
) -> Tuple:
    quantized_config = copy.deepcopy(config)
    nn.quantize(model, q_group_size, q_bits)
    quantized_config["quantization"] = {"group_size": q_group_size, "bits": q_bits}
    quantized_weights = dict(tree_flatten(model.parameters()))

    return quantized_weights, quantized_config

def convert(
    hf_path: str,
    mlx_path: str = "mlx_model",
    q_group_size: int = 64,
    q_bits: int = 4,
):
    print("[INFO] Loading")
    if isinstance(mlx_path, str):
        mlx_path = Path(mlx_path)
    model_path = get_model_path(hf_path)
    mlx_path.mkdir(parents=True, exist_ok=True)

    with open(model_path / "config.json", "r") as f:
        config = json.load(f)
    model = Nougat.from_pretrained(hf_path)
    weights = dict(tree_flatten(model.parameters()))

    print("[INFO] Quantizing")
    model.load_weights(list(weights.items()))
    weights, config = quantize_model(model, config, q_group_size, q_bits)


    for file in model_path.glob("*"):
        if file.suffix != ".safetensors":
            shutil.copy(file, mlx_path)
    del model
    save_weights(mlx_path, weights, donate_weights=True)

    save_config(config, config_path=mlx_path / "config.json")








import collections
from dataclasses import dataclass, field
import inspect
from typing import List, Optional, Tuple
import mlx.core as mx
import mlx.nn as nn

from mlx_nougat.utils import custom_roll, window_partition, window_reverse

@dataclass
class DonutSwinModelConfig:
    model_type: str = "donut-swin"
    image_size: int = 224
    patch_size: int = 4
    num_channels: int = 3
    embed_dim: int = 96
    depths: List[int] = field(default_factory=lambda: [2, 2, 6, 2])
    num_heads: List[int] = field(default_factory=lambda: [3, 6, 12, 24])
    window_size: int = 7
    mlp_ratio: float = 4.0
    qkv_bias: bool = True
    hidden_dropout_prob: float = 0.0
    attention_probs_dropout_prob: float = 0.0
    drop_path_rate: float = 0.1
    hidden_act: str = "gelu"
    use_absolute_embeddings: bool = False
    initializer_range: float = 0.02
    layer_norm_eps: float = 1e-5
    hidden_size: int = field(init=False)
    num_hidden_layers: int = field(init=False)
    chunk_size_feed_forward: int = 0

    def __post_init__(self):
        self.hidden_size = int(self.embed_dim * 2 ** (len(self.depths) - 1))
        self.num_hidden_layers = len(self.depths)

    @classmethod
    def from_dict(cls, params):
        return cls(
            **{
                k: v
                for k, v in params.items()
                if k in inspect.signature(cls).parameters
            }
        )
    
class DonutSwinPatchEmbeddings(nn.Module):
    def __init__(self, config: DonutSwinModelConfig):
        super().__init__()

        image_size, patch_size = config.image_size, config.patch_size
        num_channels, hidden_size = config.num_channels, config.embed_dim
        image_size = image_size if isinstance(image_size, collections.abc.Iterable) else (image_size, image_size)
        patch_size = patch_size if isinstance(patch_size, collections.abc.Iterable) else (patch_size, patch_size)
        num_patches = (image_size[1] // patch_size[1]) * (image_size[0] // patch_size[0])
        self.image_size = image_size
        self.patch_size = patch_size
        self.num_channels = num_channels
        self.num_patches = num_patches
        self.grid_size = (image_size[0] // patch_size[0], image_size[1] // patch_size[1])

        self.projection = nn.Conv2d(in_channels=num_channels, out_channels=hidden_size, kernel_size=patch_size, stride=patch_size, bias=True)

    def maybe_pad(self, pixel_values, height, width):
        pad_width = [(0, 0)] * 4  # Initialize pad_width with zeros for all dimensions
        if width % self.patch_size[1] != 0:
            pad_width[3] = (0, self.patch_size[1] - width % self.patch_size[1])
        if height % self.patch_size[0] != 0:
            pad_width[2] = (0, self.patch_size[0] - height % self.patch_size[0])
        pixel_values = mx.pad(pixel_values, pad_width=pad_width, constant_values=0)
        return pixel_values
    
    def __call__(
        self, pixel_values: mx.array
    ) -> mx.array:
        batch_size, height, width, num_channels = pixel_values.shape
        pixel_values = self.maybe_pad(pixel_values, height, width)
        embeddings = self.projection(pixel_values)
        batch_size, height, width, hidden_size = embeddings.shape
        embeddings = embeddings.reshape(batch_size, height * width, hidden_size)
        output_dimensions = (height, width)
        return embeddings, output_dimensions

class DonutSwinEmbeddings(nn.Module):
    def __init__(self, config: DonutSwinModelConfig):
        super().__init__()

        self.patch_embeddings = DonutSwinPatchEmbeddings(config)
        self.patch_grid = self.patch_embeddings.grid_size
        
        self.norm = nn.LayerNorm(config.embed_dim)

    def __call__(
        self, pixel_values: mx.array
    ) -> mx.array:
        embeddings, output_dimensions = self.patch_embeddings(pixel_values)
        embeddings = self.norm(embeddings)
        return embeddings, output_dimensions

class DonutSwinSelfAttention(nn.Module):
    def __init__(self, config, dim, num_heads, window_size):
            super().__init__()
            if dim % num_heads != 0:
                raise ValueError(
                    f"The hidden size ({dim}) is not a multiple of the number of attention heads ({num_heads})"
                )
            
            self.num_attention_heads = num_heads
            self.attention_head_size = int(dim / num_heads)
            self.all_head_size = self.num_attention_heads * self.attention_head_size
            self.window_size = (
                window_size if isinstance(window_size, collections.abc.Iterable) else (window_size, window_size)
            )

            self.relative_position_bias_table = mx.zeros(((2 * self.window_size[0] - 1) * (2 * self.window_size[1] - 1), num_heads))
            coords_h = mx.arange(self.window_size[0])
            coords_w = mx.arange(self.window_size[1])
            coords = mx.stack(mx.meshgrid(coords_h, coords_w, indexing='ij'), axis=0)
            coords_flatten = coords.reshape(2, -1)
            relative_coords = coords_flatten[:, :, None] - coords_flatten[:, None, :]
            relative_coords = mx.transpose(relative_coords, (1, 2, 0))
            relative_coords[:, :, 0] += self.window_size[0] - 1
            relative_coords[:, :, 1] += self.window_size[1] - 1
            relative_coords[:, :, 0] *= 2 * self.window_size[1] - 1
            self.relative_position_index = mx.sum(relative_coords, axis=-1)
            self.query = nn.Linear(self.all_head_size, self.all_head_size, bias=config.qkv_bias)
            self.key = nn.Linear(self.all_head_size, self.all_head_size, bias=config.qkv_bias)
            self.value = nn.Linear(self.all_head_size, self.all_head_size, bias=config.qkv_bias)

    def transpose_for_scores(self, x):
        new_x_shape = x.shape[:-1] + (self.num_attention_heads, self.attention_head_size)
        x = x.reshape(new_x_shape)
        return x.transpose(0, 2, 1, 3)
        
    def __call__(
        self, hidden_states: mx.array, attention_mask: Optional[mx.array] = None
    ) -> mx.array:
        batch_size, dim, num_channels = hidden_states.shape
        mixed_query_layer = self.query(hidden_states)

        key_layer = self.transpose_for_scores(self.key(hidden_states))
        value_layer = self.transpose_for_scores(self.value(hidden_states))
        query_layer = self.transpose_for_scores(mixed_query_layer)

        attention_scores = query_layer @ key_layer.transpose(0, 1, 3, 2)
        attention_scores = attention_scores / mx.sqrt(float(self.attention_head_size))

        relative_position_bias = self.relative_position_bias_table[self.relative_position_index.reshape(-1)]
        relative_position_bias = relative_position_bias.reshape(
            (self.window_size[0] * self.window_size[1], self.window_size[0] * self.window_size[1], -1))

        relative_position_bias = relative_position_bias.transpose(2, 0, 1)
        attention_scores = attention_scores + mx.expand_dims(relative_position_bias, 0)
        if attention_mask is not None:
                mask_shape = attention_mask.shape[0]
                attention_scores = attention_scores.reshape(
                    batch_size // mask_shape, mask_shape, self.num_attention_heads, dim, dim
                )
                attention_scores = attention_scores +  mx.expand_dims(mx.expand_dims(attention_mask,1),0)
                attention_scores = attention_scores.reshape(-1, self.num_attention_heads, dim, dim)

        attention_probs = mx.softmax(attention_scores.astype(mx.float32), axis=-1).astype(attention_scores.dtype)
        context_layer = attention_probs @ value_layer
        context_layer = context_layer.transpose(0, 2, 1, 3)
        new_context_layer_shape = context_layer.shape[:-2] + (self.all_head_size,)
        context_layer = context_layer.reshape(new_context_layer_shape)

        return (context_layer,)           
 
class DonutSwinSelfOutput(nn.Module):
    def __init__(self, config, dim):
        super().__init__()
        self.dense = nn.Linear(dim, dim)

    def __call__(
        self, hidden_states: mx.array, input_tensor: mx.array
    ) -> mx.array:
        hidden_states = self.dense(hidden_states)
        return hidden_states
    
class DonutSwinAttention(nn.Module):
    def __init__(self, config, dim, num_heads, window_size):
        super().__init__()
        self.self = DonutSwinSelfAttention(config, dim, num_heads, window_size)
        self.output = DonutSwinSelfOutput(config, dim)

    def __call__(
        self, hidden_states: mx.array, attention_mask: Optional[mx.array] = None
    ) -> mx.array:
        self_outputs = self.self(hidden_states, attention_mask)
        attention_output = self.output(self_outputs[0], hidden_states)
        return (attention_output,) + self_outputs[1:]

class DonutSwinIntermediate(nn.Module):
    def __init__(self, config, dim):
        super().__init__()
        self.dense = nn.Linear(dim, int(config.mlp_ratio * dim))
        self.intermediate_act_fn = nn.GELU()

    def __call__(
        self, hidden_states: mx.array
    ) -> mx.array:
        hidden_states = self.dense(hidden_states)
        hidden_states = self.intermediate_act_fn(hidden_states)
        return hidden_states
    
class DonutSwinOutput(nn.Module):
    def __init__(self, config, dim):
        super().__init__()
        self.dense = nn.Linear(int(config.mlp_ratio * dim), dim)

    def __call__(
        self, hidden_states: mx.array, 
    ) -> mx.array:
        hidden_states = self.dense(hidden_states)
        return hidden_states    

class DonutSwinLayer(nn.Module):
    def __init__(self, config: DonutSwinModelConfig, dim: int, input_resolution: tuple, num_heads: int, shift_size: int):
        super().__init__()
        self.chunk_size_feed_forward = config.chunk_size_feed_forward
        self.shift_size = shift_size
        self.window_size = config.window_size
        self.input_resolution = input_resolution
        self.layernorm_before = nn.LayerNorm(dim, eps=config.layer_norm_eps)
        self.attention = DonutSwinAttention(config, dim, num_heads, window_size=self.window_size)
        self.layernorm_after = nn.LayerNorm(dim, eps=config.layer_norm_eps)
        self.intermediate = DonutSwinIntermediate(config, dim)
        self.output = DonutSwinOutput(config, dim)

    def set_shift_and_window_size(self, input_resolution):
        if min(input_resolution) <= self.window_size:
            # if window size is larger than input resolution, we don't partition windows
            self.shift_size = 0
            self.window_size = min(input_resolution)
    def maybe_pad(self, hidden_states, height, width):
            pad_right = (self.window_size - width % self.window_size) % self.window_size
            pad_bottom = (self.window_size - height % self.window_size) % self.window_size
            pad_values = [(0, 0), (0, pad_bottom), (0, pad_right), (0,0)]  
            hidden_states = mx.pad(hidden_states, pad_width=pad_values, constant_values=0)
            return hidden_states, pad_values
    
    def get_attn_mask(self, height, width):
        if self.shift_size > 0:
            img_mask = mx.zeros((1, height, width, 1))
            height_slices = (
                slice(0, -self.window_size),
                slice(-self.window_size, -self.shift_size),
                slice(-self.shift_size, None),
            )
            width_slices = (
                slice(0, -self.window_size),
                slice(-self.window_size, -self.shift_size),
                slice(-self.shift_size, None),
            )
            count = 0
            for height_slice in height_slices:
                for width_slice in width_slices:
                    img_mask[:, height_slice, width_slice, :] = count
                    count += 1
            mask_windows = window_partition(img_mask, self.window_size)
            mask_windows = mask_windows.reshape(-1, self.window_size * self.window_size)
            attn_mask = mask_windows[:, mx.newaxis, :] - mask_windows[:, :, mx.newaxis]
            attn_mask = mx.where(attn_mask != 0, -100.0, 0.0).astype(mx.float32)
        else:
            attn_mask = None
        return attn_mask    
    
    def __call__(
        self, hidden_states: mx.array, input_dimensions: Tuple[int, int], attention_mask: Optional[mx.array] = None
    ) -> mx.array:
        self.set_shift_and_window_size(input_dimensions)
        height, width = input_dimensions
        batch_size, _, channels = hidden_states.shape
        shortcut = hidden_states
        hidden_states = self.layernorm_before(hidden_states)

        hidden_states = hidden_states.reshape(batch_size, height, width, channels)

        hidden_states, pad_values = self.maybe_pad(hidden_states, height, width)
        _, height_pad, width_pad, _ = hidden_states.shape
        # cyclic shift
        if self.shift_size > 0:
            shifted_hidden_states = custom_roll(hidden_states, shift=(-self.shift_size, -self.shift_size), axis=(1, 2))
        else:
            shifted_hidden_states = hidden_states

        hidden_states_windows = window_partition(shifted_hidden_states, self.window_size)
        hidden_states_windows = hidden_states_windows.reshape(-1, self.window_size * self.window_size, channels)
        attn_mask = self.get_attn_mask(
            height_pad, width_pad
        )
        attention_outputs = self.attention(
            hidden_states_windows, attn_mask
        )

        attention_output = attention_outputs[0]
        attention_windows = attention_output.reshape(-1, self.window_size, self.window_size, channels)
        shifted_windows = window_reverse(attention_windows, self.window_size, height_pad, width_pad)

        if self.shift_size > 0:
            attention_windows = custom_roll(shifted_windows, shift=(self.shift_size, self.shift_size), axis=(1, 2))
        else:
            attention_windows = shifted_windows

        was_padded = pad_values[1][1] > 0 or pad_values[2][1] > 0
        if was_padded:
            attention_windows = attention_windows[:, :height, :width, :]

        attention_windows = attention_windows.reshape(batch_size, height * width, channels)

        hidden_states = shortcut + attention_windows
        layer_output = self.layernorm_after(hidden_states)
        layer_output = self.intermediate(layer_output)
        layer_output = hidden_states + self.output(layer_output)

        return layer_output

class DonutSwinStage(nn.Module):
    def __init__(
        self,
        config: DonutSwinModelConfig,
        dim: int,
        input_resolution: tuple,
        depth: int,
        num_heads: int,
        drop_path: List[float],
        downsample: Optional[nn.Module] = None,
    ):
        super().__init__()
        self.config = config
        self.dim = dim
        self.blocks = [
                DonutSwinLayer(
                    config=config,
                    dim=dim,
                    input_resolution=input_resolution,
                    num_heads=num_heads,
                    shift_size=0 if (i % 2 == 0) else config.window_size // 2,
                )
            for i in range(depth)
        ]

        if downsample is not None:
            self.downsample = downsample(input_resolution, dim=dim, norm_layer=nn.LayerNorm)
        else:
            self.downsample = None

        self.pointing = False
    def __call__(
        self, hidden_states: mx.array, input_dimensions: Tuple[int, int], head_mask: Optional[mx.array] = None
    ) -> mx.array:
        height, width = input_dimensions
        for i, layer_module in enumerate(self.blocks):
            layer_head_mask = head_mask[i] if head_mask is not None else None

            hidden_states = layer_module(
                hidden_states, input_dimensions, layer_head_mask
            )

        hidden_states_before_downsampling = hidden_states

        if self.downsample is not None:
            height_downsampled, width_downsampled = (height + 1) // 2, (width + 1) // 2
            output_dimensions = (height, width, height_downsampled, width_downsampled)
            hidden_states = self.downsample(hidden_states_before_downsampling, input_dimensions)
        else:
            output_dimensions = (height, width, height, width)

        stage_outputs = (hidden_states, hidden_states_before_downsampling, output_dimensions)

        return stage_outputs

class DonutSwinPatchMerging(nn.Module):
    def __init__(self, input_resolution: Tuple[int], dim: int, norm_layer: nn.Module = nn.LayerNorm) -> None:
        super().__init__()
        self.input_resolution = input_resolution
        self.dim = dim
        self.reduction = nn.Linear(4 * dim, 2 * dim, bias=False)
        self.norm = norm_layer(4 * dim)
    def maybe_pad(self, input_feature, height, width):
        should_pad = (height % 2 == 1) or (width % 2 == 1)
        if should_pad:
            pad_height = height % 2
            pad_width = width % 2
            pad_values = [(0, 0), (0, pad_height), (0, pad_width), (0, 0)]
            input_feature = mx.pad(input_feature, pad_width=pad_values, constant_values=0)

        return input_feature
    
    def __call__(self, input_feature: mx.array, input_dimensions: Tuple[int, int]) -> mx.array:
        height, width = input_dimensions
        # `dim` is height * width
        batch_size, dim, num_channels = input_feature.shape

        input_feature = input_feature.reshape(batch_size, height, width, num_channels)
        # pad input to be divisible by width and height, if needed
        input_feature = self.maybe_pad(input_feature, height, width)
        # [batch_size, height/2, width/2, num_channels]
        input_feature_0 = input_feature[:, 0::2, 0::2, :]
        # [batch_size, height/2, width/2, num_channels]
        input_feature_1 = input_feature[:, 1::2, 0::2, :]
        # [batch_size, height/2, width/2, num_channels]
        input_feature_2 = input_feature[:, 0::2, 1::2, :]
        # [batch_size, height/2, width/2, num_channels]
        input_feature_3 = input_feature[:, 1::2, 1::2, :]
        # batch_size height/2 width/2 4*num_channels
        input_feature = mx.concatenate([input_feature_0, input_feature_1, input_feature_2, input_feature_3], axis=-1)
        input_feature = input_feature.reshape(batch_size, -1, 4 * num_channels)  # batch_size height/2*width/2 4*C

        input_feature = self.norm(input_feature)
        input_feature = self.reduction(input_feature)

        return input_feature
          
class DonutSwinEncoder(nn.Module):
    def __init__(self, config: DonutSwinModelConfig, grid_size):
        super().__init__()

        self.config = config
        self.num_layers = len(config.depths)
        dpr = mx.linspace(start=0, stop=config.drop_path_rate, num=sum(config.depths)).tolist()
        self.layers =[
            DonutSwinStage(
                    config=config,
                    dim=int(config.embed_dim * 2**layer_id),
                    input_resolution=(grid_size[0] // (2**layer_id), grid_size[1] // (2**layer_id)),
                    depth=config.depths[layer_id],
                    num_heads=config.num_heads[layer_id],
                    drop_path=dpr[sum(config.depths[:layer_id]) : sum(config.depths[: layer_id + 1])],
                    downsample=DonutSwinPatchMerging if (layer_id < self.num_layers - 1) else None,
                ) for layer_id in range(self.num_layers)
            ]
    def __call__(
        self, hidden_states: mx.array, input_dimensions: Tuple[int, int], head_mask: Optional[mx.array] = None
    ) -> mx.array:
        hidden_states = hidden_states
        for i, layer_module in enumerate(self.layers):
            layer_head_mask = head_mask[i] if head_mask is not None else None
            layer_outputs = layer_module(hidden_states, input_dimensions, layer_head_mask)
            hidden_states = layer_outputs[0]
            output_dimensions = layer_outputs[2]
            input_dimensions = (output_dimensions[-2], output_dimensions[-1])

        return hidden_states
    
class DonutSwinModel(nn.Module):
    def __init__(self, config: DonutSwinModelConfig):
        super().__init__()

        self.config = config
        self.num_layers = len(config.depths)
        self.num_features = int(config.embed_dim * 2 ** (self.num_layers - 1))

        self.embeddings = DonutSwinEmbeddings(config)
        self.encoder = DonutSwinEncoder(config, self.embeddings.patch_grid)

    def __call__(
        self, pixel_values: mx.array, head_mask: Optional[mx.array] = None
    ) -> mx.array:
        if pixel_values is None:
            raise ValueError("You have to specify pixel_values")
        embedding_output, input_dimensions = self.embeddings(
            pixel_values
        )
        encoder_outputs = self.encoder(embedding_output,input_dimensions,head_mask=head_mask)
        return encoder_outputs
    @staticmethod
    def sanitize(weights):
        sanitized_weights = {}
        for k, v in weights.items():
            if k.startswith("decoder."):
                continue
            # remove `encoder.` prefix in k
            k = k.replace("encoder.", "", 1)
            if "projection.weight" in k:
                # PyTorch conv2d weight tensors have shape:
                #   [out_channels, in_channels, kH, KW]
                # MLX conv2d expects the weight be of shape:
                #   [out_channels, kH, KW, in_channels]
                sanitized_weights[k] = v.transpose(0, 2, 3, 1)
            else:
                sanitized_weights[k] = v
        return sanitized_weights


from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="mlx_nougat",
    version="0.1.2",
    author="anchen",
    author_email="li.anchen.au@gmail.com",
    description="A CLI tool for OCR using the Nougat model",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mzbac/mlx_nougat",
    packages=find_packages(),
    install_requires=[
        "numpy",
        "requests",
        "pillow",
        "transformers",
        "wand",
        "mlx",
    ],
    entry_points={
        "console_scripts": [
            "mlx_nougat=mlx_nougat.cli:main",
        ],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)
